.686
.model flat,stdcall
option casemap:none

include \masm32\include\w2k\ntstatus.inc
include \masm32\include\w2k\ntddk.inc
include \masm32\include\w2k\w2kundoc.inc

include \masm32\macros\strings.asm

include \masm32\include\w2k\ntoskrnl.inc
includelib \masm32\lib\w2k\ntoskrnl.lib

;Control Codes
IOCTL_VIRTTOPHYS equ CTL_CODE(FILE_DEVICE_UNKNOWN,800h,METHOD_BUFFERED,FILE_READ_DATA or FILE_WRITE_DATA)
IOCTL_CHANGEPHYS equ CTL_CODE(FILE_DEVICE_UNKNOWN,801h,METHOD_BUFFERED,FILE_READ_DATA or FILE_WRITE_DATA)

;Function Prototypes
ChangePhys proto :DWORD,:DWORD
VirtToPhys proto :DWORD

GetPTEAddress proto :DWORD

.const

;Device Name Strings
CCOUNTED_UNICODE_STRING "\\Device\\MemMgr",usDevice,4
CCOUNTED_UNICODE_STRING "\\??\\MemMgr",usSymbolicLink,4

.data

.code

;----------------------------------------------------------------
;DRIVER MANAGEMENT ROUTINES
;----------------------------------------------------------------

;----------------------------------------------------------------
;DriverEntry
;----------------------------------------------------------------
;Called by Windows when the driver is started
;----------------------------------------------------------------

DriverEntry proc pDriverObject:DWORD,pRegPath:DWORD
	
	LOCAL pDeviceObject:DWORD
	
	;Create the device and a symbolic link
	
	invoke IoCreateDevice,pDriverObject,0,addr usDevice,FILE_DEVICE_UNKNOWN,0,FALSE,addr pDeviceObject
	.if eax==STATUS_SUCCESS
		invoke IoCreateSymbolicLink,addr usSymbolicLink,addr usDevice
		.if eax==STATUS_SUCCESS
			;Tell Windows what functions to call
			mov eax,pDriverObject
			assume eax:ptr DRIVER_OBJECT
			mov [eax].MajorFunction+IRP_MJ_CREATE*4,offset DispatchCreate
			mov [eax].MajorFunction+IRP_MJ_CLOSE*4,offset DispatchClose
			mov [eax].MajorFunction+IRP_MJ_DEVICE_CONTROL*4,offset DispatchControl
			mov [eax].DriverUnload,offset DriverUnload
			assume eax:nothing
			mov eax,STATUS_SUCCESS
		.else
			invoke IoDeleteDevice,pDeviceObject
			mov eax,STATUS_DEVICE_CONFIGURATION_ERROR
		.endif
	.else
		mov eax,STATUS_DEVICE_CONFIGURATION_ERROR
	.endif
	ret

DriverEntry endp

;----------------------------------------------------------------
;DriverUnload
;----------------------------------------------------------------
;Called by Windows when the driver is stopped
;----------------------------------------------------------------

DriverUnload proc pDriverObject:DWORD

	;Delete the device and symbolic link
	
	invoke IoDeleteSymbolicLink,addr usSymbolicLink
	mov eax,pDriverObject
	assume eax:ptr DRIVER_OBJECT
	invoke IoDeleteDevice,[eax].DeviceObject
	assume eax:nothing
	ret

DriverUnload endp

;----------------------------------------------------------------
;DispatchCreate
;----------------------------------------------------------------
;Called by Windows when a handle to the device is opened
;----------------------------------------------------------------

DispatchCreate proc pDeviceObject:DWORD,pIRP:DWORD

	;Return STATUS_SUCCESS
	
	mov eax,pIRP
	assume eax:ptr _IRP
	mov [eax].IoStatus.Status,STATUS_SUCCESS
	mov [eax].IoStatus.Information,0
	assume eax:nothing
	invoke IoCompleteRequest,pIRP,IO_NO_INCREMENT
	mov eax,STATUS_SUCCESS
	ret

DispatchCreate endp

;----------------------------------------------------------------
;DispatchControl
;----------------------------------------------------------------
;Called by Windows when all handles to the device are closed
;----------------------------------------------------------------

DispatchClose proc pDeviceObject:DWORD,pIRP:DWORD

	;Return STATUS_SUCCESS
	
	mov eax,pIRP
	assume eax:ptr _IRP
	mov [eax].IoStatus.Status,STATUS_SUCCESS
	mov [eax].IoStatus.Information,0
	assume eax:nothing
	invoke IoCompleteRequest,pIRP,IO_NO_INCREMENT
	mov eax,STATUS_SUCCESS
	ret

DispatchClose endp

;----------------------------------------------------------------
;DispatchControl
;----------------------------------------------------------------
;Called by Windows when DeviceIoControl is used on the device
;----------------------------------------------------------------

DispatchControl proc pDeviceObject:DWORD,pIRP:DWORD

	;Find out which control code was passed
	;Call the associated function
	
	mov edx,pIRP
	assume edx:ptr _IRP
	mov eax,[edx].Tail.Overlay.CurrentStackLocation
	assume eax:ptr IO_STACK_LOCATION
	
	mov ecx,[eax].Parameters.DeviceIoControl.IoControlCode
	.if ecx==IOCTL_VIRTTOPHYS
		.if [eax].Parameters.DeviceIoControl.InputBufferLength==4 && [eax].Parameters.DeviceIoControl.OutputBufferLength==4
			mov ecx,[edx].AssociatedIrp.SystemBuffer
			mov eax,[ecx]
			push ecx
			invoke VirtToPhys,eax
			pop ecx
			mov [ecx],eax
			mov edx,pIRP
			mov [edx].IoStatus.Status,STATUS_SUCCESS
			mov [edx].IoStatus.Information,4
		.else
			mov [edx].IoStatus.Status,STATUS_INVALID_BUFFER_SIZE
			mov [edx].IoStatus.Information,0
		.endif
	.elseif ecx==IOCTL_CHANGEPHYS
		.if [eax].Parameters.DeviceIoControl.InputBufferLength==8 && [eax].Parameters.DeviceIoControl.OutputBufferLength==8
			mov ecx,[edx].AssociatedIrp.SystemBuffer
			mov eax,[ecx]
			mov edx,[ecx+4]
			push ecx
			invoke ChangePhys,eax,edx
			pop ecx
			mov [ecx+4],eax
			mov edx,pIRP
			mov [edx].IoStatus.Status,STATUS_SUCCESS
			mov [edx].IoStatus.Information,8
		.else
			mov [edx].IoStatus.Status,STATUS_INVALID_BUFFER_SIZE
			mov [edx].IoStatus.Information,0
		.endif
	.else
		mov [edx].IoStatus.Status,STATUS_INVALID_DEVICE_REQUEST
		mov [edx].IoStatus.Information,0
	.endif
	invoke IoCompleteRequest,pIRP,IO_NO_INCREMENT
	assume eax:nothing
	assume edx:nothing
	ret

DispatchControl endp

;----------------------------------------------------------------
;EXPORTED DRIVER FUNCTIONS
;----------------------------------------------------------------

;----------------------------------------------------------------
;VirtToPhys
;----------------------------------------------------------------
;Translates a virtual address to a physical address
;Performs the same action as MmGetPhysicalAddress
;----------------------------------------------------------------

VirtToPhys proc pVirtual:DWORD
	
	invoke GetPTEAddress,pVirtual
	.if eax!=0
		;eax = offset PTE
		mov eax,[eax]
		;eax = PTE
		and eax,0FFFFF000h
		;eax = upper 20 bits of PTE
		;eax = physical address pointed to by PTE
		mov edx,pVirtual
		and edx,000000FFFh
		;edx = lower 12 bits of virtual address
		;edx = offset into page
		add eax,edx
		;base + offset
	.endif
	ret

VirtToPhys endp

;----------------------------------------------------------------
;ChangePhys
;----------------------------------------------------------------
;Makes the virtual address given in pVirtual translate to the
;physical address in pPhysical.
;Returns the old physical address pointed to by pVirtual
;The lower 12 bits of all addresses are ignored.#
;----------------------------------------------------------------

ChangePhys proc pVirtual:DWORD,pPhysical:DWORD
	
	invoke GetPTEAddress,pVirtual
	.if eax!=0
		;eax = &PTE
		mov edx,[eax]
		mov ecx,edx
		;ecx = edx = PTE
		and ecx,0FFFFF000h
		;ecx = upper 20 bits of PTE
		;ecx = old physical address
		push ecx
		and edx,000000FFFh
		;edx = lower 12 bits of PTE
		;edx = old PTE flags
		mov ecx,pPhysical
		and ecx,0FFFFF000h
		;ecx = new physical address
		mov [eax],ecx	;insert new physical address
		add [eax],edx	;insert old flags
		pop eax
		;return value = old physical address
	.endif
	ret

ChangePhys endp

;----------------------------------------------------------------
;INTERNAL DRIVER FUNCTIONS
;----------------------------------------------------------------

;----------------------------------------------------------------
;GetPTEAddress
;----------------------------------------------------------------
;Finds the PTE (PDE for 4MB pages) for a specified virtual
;address and returns its virtual address.
;Returns 0 if the page is not committed.
;----------------------------------------------------------------

GetPTEAddress proc pVirtual:DWORD
	
	mov eax,pVirtual
	shr eax,22
	shl eax,2
	;eax = (upper 10 bits of pVirtual) * sizeof dword
	;eax = PDE offset into PD
	add eax,0C0300000h
	;eax = address of PDE
	mov edx,[eax]
	.if (edx & 00000001b)
		;page is committed
		.if !(edx & 10000000b)
			;4k pages
			mov eax,pVirtual
			shr eax,12
			shl eax,2
			;eax = (upper 20 bits of pVirtual) * sizeof dword
			;eax = PTE offset into PT array
			add eax,0C0000000h
			;eax = address of PTE
			mov edx,[eax]
			.if !(edx & 00000001b)
				;page is not committed
				xor eax,eax
			.endif
		.endif
		;if not 4k, then pages are 4Mb therefore PTE = PDE
	.else
		;page is not committed
		xor eax,eax
	.endif
	ret

GetPTEAddress endp

end DriverEntry